{# Import the theme's layout. #} {% extends "!layout.html" %} {# Custom CSS overrides #} {% set bootswatch_css_custom = ['_static/pagmo.css']
%} {% block extrahead %}
<link rel="stylesheet" type="text/css" href="https://fonts.googleapis.com/css?family=Roboto|Comfortaa|Space+Mono"> {% endblock %}